package core

import (
	"math"

	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/eval"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// Evaluate evaluates an expression and returns a register that contains the value of the expression.
func (f *Function) Evaluate(expr *expression.Expression) (eval.Value, error) {
	if expr.IsLeaf() {
		if expr.Token.IsNumeric() {
			number, err := f.ToNumber(expr.Token)

			if err != nil {
				return eval.Value{}, err
			}

			value := eval.Value{
				Kind:   eval.Number,
				Type:   types.AnyInt,
				Number: number,
				Alive:  1,
			}

			return value, nil
		}

		if expr.Token.Kind == token.Identifier {
			name := expr.Token.Text(f.File.Bytes)
			variable, function := f.Identifier(name)

			if variable != nil {
				f.UseVariable(variable)

				if variable.Alive == 0 {
					return variable.Value, nil
				}

				tmp := f.NewRegister()
				f.RegisterRegister(asm.MOVE, tmp, variable.Register)

				value := eval.Value{
					Kind:     eval.Register,
					Type:     variable.Type,
					Register: tmp,
					Alive:    1,
				}

				return value, nil
			}

			if function != nil {
				value := eval.Value{
					Kind:  eval.Label,
					Type:  types.AnyPointer,
					Label: function.UniqueName,
					Alive: 1,
				}

				return value, nil
			}

			return eval.Value{}, errors.New(&errors.UnknownIdentifier{Name: name}, f.File, expr.Token.Position)
		}
	}

	if expr.Token.Kind == token.Call {
		types, err := f.CompileCall(expr)

		if err != nil {
			return eval.Value{}, err
		}

		if len(types) == 0 {
			return eval.Value{}, errors.New(errors.UntypedExpression, f.File, expr.Token.Position)
		}

		value := eval.Value{
			Kind:     eval.Register,
			Type:     types[0],
			Register: f.CPU.Output[0],
			Alive:    1,
		}

		return value, nil
	}

	if expr.Token.Kind == token.Period {
		left := expr.Children[0]
		right := expr.Children[1]
		leftText := left.Token.Text(f.File.Bytes)
		rightText := right.Token.Text(f.File.Bytes)
		variable := f.VariableByName(leftText)

		if variable != nil {
			field := variable.Type.(*types.Pointer).To.(*types.Struct).FieldByName(rightText)

			value := eval.Value{
				Kind:  eval.Memory,
				Type:  field.Type,
				Alive: 1,
				Memory: asm.Memory{
					Base:           variable.Register,
					Offset:         int8(field.Offset),
					OffsetRegister: math.MaxUint8,
					Length:         byte(field.Type.Size()),
				},
			}

			return value, nil
		}

		constant, isConst := f.All.Constants[f.Package+"."+leftText+"."+rightText]

		if isConst {
			number, err := ToNumber(constant.Token, constant.File)

			if err != nil {
				return eval.Value{}, err
			}

			value := eval.Value{
				Kind:   eval.Number,
				Type:   types.AnyInt,
				Number: number,
				Alive:  1,
			}

			return value, nil
		}
	}

	tmp := f.NewRegister()
	typ, err := f.ExpressionToRegister(expr, tmp)

	value := eval.Value{
		Kind:     eval.Register,
		Type:     typ,
		Register: tmp,
		Alive:    1,
	}

	return value, err
}
